##############################################################################
# ScoDoc
# Copyright (c) 1999 - 2024 Emmanuel Viennet.  All rights reserved.
# See LICENSE
##############################################################################

"""
ScoDoc 9 API : accès aux demandes d'inscriptions


CATEGORY
--------
FormSemestreDemandeInscription

"""
import base64

from flask import g, make_response, request
from flask_json import as_json
from flask_login import login_required

from app import db, log
from app.api import api_bp as bp, api_web_bp, image_response
from app.api import api_permission_required as permission_required
from app.decorators import scodoc
from app.scodoc.sco_permissions import Permission
from app.scodoc import sco_utils as scu
from app.scodoc.sco_utils import json_error

from app.models import (
    FormSemestre,
    FormSemestreDemandeInscription,
    StatutDemandeInscription,
)


@bp.route(
    "/formsemestre/<int:formsemestre_id>/demande_inscription/create", methods=["POST"]
)
@api_web_bp.route(
    "/formsemestre/<int:formsemestre_id>/demande_inscription/create", methods=["POST"]
)
@login_required
@scodoc
@permission_required(Permission.DemandeInscription)
@as_json
def formsemestre_demande_inscription_create(formsemestre_id: int):
    """
    Dépôt d'une demande d'inscription à un formsemestre
    La pièce jointe images peut être passée dans le json, encodée en base64.


    formsemestre_id : l'id du formsemestre

    DATA
    ----
    ```json
    {
        'code_nip' : 'NIP', /* optionnel */
        'code_ine' : 'INE', /* optionnel */
        'etape' : 'V1RT', /* optionnel, étape principale d'inscription pour les étudiants locaux */
        'civilite' : 'X', 'M' ou 'F',
        'nom': 'MonNom',
        'prenom': 'MonPrénom',
        'email': 'nom.prenom@qq.part',
        'telephone': '0102030405', 'niveau': 'intermediaire',
        'motivation': "J'aime la langue",
        "piece_jointe" : image_en_base64,
    }
    ```
    """
    formsemestre = FormSemestre.get_formsemestre(formsemestre_id)
    # Vérification ouverture inscriptions: les dates
    # La capacité d'accueil n'est pas vérifiée ici, mais le sera au moment
    # de la validation de la demande et de l'inscription effective.
    if (
        not formsemestre.description
        or not formsemestre.description.inscriptions_ouvertes()
    ):
        return json_error(
            400,
            "Les inscriptions à ce formsemestre ne sont pas ouvertes.",
        )
    #
    args = request.get_json(force=True)  # may raise 400 Bad Request
    for key in ("nom", "prenom", "email"):
        if key not in args or not args[key].strip():
            return json_error(400, f"Le champ '{key}' est obligatoire.")
    if "piece_jointe" in args and args["piece_jointe"] is not None:
        try:
            args["piece_jointe"] = base64.b64decode(args["piece_jointe"])
        except (TypeError, ValueError, base64.binascii.Error):
            return json_error(400, "Erreur dans le format de la pièce jointe.")

    args["formsemestre_id"] = formsemestre.id
    if "formsemestre" in args:
        del args["formsemestre"]
    demande = FormSemestreDemandeInscription.create(
        civilite=args.get("civilite", "X"),
        nom=args["nom"],
        prenom=args["prenom"],
        email=args["email"],
        telephone=args.get("telephone", ""),
        code_nip=args.get("code_nip", ""),
        code_ine=args.get("code_ine", ""),
        etape=args.get("etape", None),
        niveau=args.get("niveau", ""),
        motivation=args.get("motivation", ""),
        piece_jointe=args.get("piece_jointe", None),
        formsemestre_id=formsemestre.id,
    )
    db.session.add(demande)
    db.session.commit()
    db.session.refresh(demande)
    demande.notify_demande()
    demande.process_auto()
    return demande.to_dict(exclude_images=True)


@bp.route(
    "/formsemestre/<int:formsemestre_id>/demande_inscription/<string:demande_num>/edit",
    methods=["POST"],
)
@api_web_bp.route(
    "/formsemestre/<int:formsemestre_id>/demande_inscription/<string:demande_num>/edit",
    methods=["POST"],
)
@login_required
@scodoc
@permission_required(Permission.DemandeInscription)
@as_json
def formsemestre_demande_inscription_edit(formsemestre_id: int, demande_num: str):
    """Edit une demande d'inscription (y compris piece jointe)"""
    formsemestre = FormSemestre.get_formsemestre(formsemestre_id)
    demande = FormSemestreDemandeInscription.query.filter_by(numero=demande_num).first()
    if not demande:
        return json_error(404, "Demande d'inscription non trouvée.")
    if demande.formsemestre_id != formsemestre.id:
        return json_error(400, "Demande d'inscription non trouvée dans ce semestre.")
    args = request.get_json(force=True)  # may raise 400 Bad Request
    if "piece_jointe" in args and args["piece_jointe"] is not None:
        args["piece_jointe"] = base64.b64decode(args["piece_jointe"])
    demande.from_dict(args, log=True)
    db.session.commit()
    db.session.refresh(demande)
    return demande.to_dict()


@bp.route(
    "/formsemestre/<int:formsemestre_id>/demande_inscription/<string:demande_num>/delete",
    methods=["POST", "DELETE"],
)
@api_web_bp.route(
    "/formsemestre/<int:formsemestre_id>/demande_inscription/<string:demande_num>/delete",
    methods=["POST", "DELETE"],
)
@login_required
@scodoc
@permission_required(Permission.DemandeInscription)
@as_json
def formsemestre_demande_inscription_delete(formsemestre_id: int, demande_num: str):
    """
    Suppression d'une demande d'inscription à un formsemestre
    La demande est identifiée par son numéro.
    """
    formsemestre = FormSemestre.get_formsemestre(formsemestre_id)
    demande = formsemestre.demandes_inscriptions.filter_by(
        numero=demande_num
    ).first_or_404()
    demande.delete()
    db.session.commit()
    return ("ok", 204)


@bp.route(
    "/demande_inscription/<string:demande_num>/cancel",
    methods=["POST"],
)
@api_web_bp.route(
    "/demande_inscription/<string:demande_num>/cancel",
    methods=["POST"],
)
@login_required
@scodoc
@permission_required(Permission.DemandeInscription)
@as_json
def demande_inscription_cancel(demande_num: str):
    """
    Suppression d'une demande d'inscription à un formsemestre
    La demande est identifiée par son numéro.
    """
    demande: FormSemestreDemandeInscription = (
        FormSemestreDemandeInscription.query.filter_by(
            numero=demande_num
        ).first_or_404()
    )
    if g.scodoc_dept is not None:
        # Vérifie departement
        if demande.formsemestre.dept_id != g.scodoc_dept_id:
            return json_error(
                400, "Demande d'inscription non trouvée dans ce semestre."
            )
    if demande.statut != StatutDemandeInscription.ANNULEE:
        demande.statut = StatutDemandeInscription.ANNULEE
        db.session.add(demande)
        db.session.commit()
        log(f"demande_inscription_cancel({demande.numero})")
    return demande.to_dict()


@bp.route(
    "/formsemestre/<int:formsemestre_id>/demande_inscription/accept", methods=["POST"]
)
@api_web_bp.route("/formsemestre/<int:formsemestre_id>/demande_inscription/accept")
@login_required
@scodoc
@permission_required(Permission.EtudInscrit)
@permission_required(Permission.DemandeInscription)
@as_json
def formsemestre_demande_inscription_accept(formsemestre_id: int, demande_num: str):
    """
    Validation d'une demande d'inscription à un formsemestre
    La demande est identifiée par son numéro.
    La demande passera à l'état ACCEPTEE et l'étudiant sera inscrit au formsemestre.

    Il faut la permission `DemandeInscription` et `EtudInscrit`.
    """
    formsemestre = FormSemestre.get_formsemestre(formsemestre_id)
    demande: FormSemestreDemandeInscription = (
        formsemestre.demandes_inscriptions.filter_by(numero=demande_num).first_or_404()
    )
    demande.accept()
    db.session.commit()
    db.session.refresh(demande)
    return demande.to_dict()


@bp.route(
    "/formsemestre/<int:formsemestre_id>/demande_inscription/<string:demande_num>"
)
@api_web_bp.route(
    "/formsemestre/<int:formsemestre_id>/demande_inscription/<string:demande_num>"
)
@login_required
@scodoc
@permission_required(Permission.DemandeInscription)
@as_json
def formsemestre_demande_inscription_get(formsemestre_id: int, demande_num: str):
    """
    La demande d'inscription à un formsemestre, identifiée par son numéro.
    """
    formsemestre = FormSemestre.get_formsemestre(formsemestre_id)
    demande = FormSemestreDemandeInscription.query.filter_by(numero=demande_num).first()
    if not demande:
        return json_error(404, "Demande d'inscription non trouvée.")
    if demande.formsemestre_id != formsemestre.id:
        return json_error(400, "Demande d'inscription non trouvée dans ce semestre.")
    d = demande.to_dict(exclude_images=False)
    if "piece_jointe" in d and d["piece_jointe"] is not None:
        d["piece_jointe"] = base64.b64encode(d["piece_jointe"]).decode("utf-8")
    return d


@bp.route("/demandes_inscriptions/by_mail/<string:email>")
@api_web_bp.route("/demandes_inscriptions/by_mail/<string:email>")
@login_required
@scodoc
@permission_required(Permission.DemandeInscription)
@as_json
def demandes_inscriptions_get_by_mail(email: str):
    """
    Liste des demandes d'inscription concernant cet apprenant, identifiées par son email.
    (sans les pièces jointes)
    """
    demandes = FormSemestreDemandeInscription.query.filter_by(email=email)
    if g.scodoc_dept is not None:
        # Si route départementale, élimine les demandes d'autres départements
        demandes = [
            demande
            for demande in demandes
            if demande.formsemestre.dept_id == g.scodoc_dept_id
        ]

    return [demande.to_dict(exclude_images=True) for demande in demandes]


@bp.route("/demande_inscription/<string:demande_num>")
@api_web_bp.route("/demande_inscription/<string:demande_num>")
@login_required
@scodoc
@permission_required(Permission.DemandeInscription)
@as_json
def demande_inscription_get(demande_num: str):
    """
    La demande d'inscription identifiée par son numéro.
    """
    demande: FormSemestreDemandeInscription = (
        FormSemestreDemandeInscription.query.filter_by(numero=demande_num).first()
    )
    if not demande:
        return json_error(404, "Demande d'inscription non trouvée.")
    formsemestre = demande.formsemestre
    if g.scodoc_dept is not None:
        # Vérifie departement
        if formsemestre.dept_id != g.scodoc_dept_id:
            return json_error(
                400, "Demande d'inscription non trouvée dans ce semestre."
            )
    d = demande.to_dict(exclude_images=False)
    if "piece_jointe" in d and d["piece_jointe"] is not None:
        d["piece_jointe"] = base64.b64encode(d["piece_jointe"]).decode("utf-8")
    return d


@bp.route("demande_inscription/<string:demande_num>/piece_jointe")
@api_web_bp.route("demande_inscription/<string:demande_num>/piece_jointe")
@login_required
@scodoc
@permission_required(Permission.DemandeInscription)
def demande_inscription_get_piece_jointe(demande_num: str):
    """
    La pièce jointe de la demande d'inscription à un formsemestre, identifiée par son numéro.
    La réponse est vide (204) si pas de pièce attachée.
    Sinon, c'est une image (jpeg, png, ou pdf) avec le bon type MIME.
    """
    demande = FormSemestreDemandeInscription.query.filter_by(numero=demande_num).first()
    if not demande:
        return json_error(404, "Demande d'inscription non trouvée.")
    if g.scodoc_dept is not None:
        # Vérifie departement
        if demande.formsemestre.dept_id != g.scodoc_dept_id:
            return json_error(
                400, "Demande d'inscription non trouvée dans ce semestre."
            )
    if not demande.piece_jointe:
        return make_response("", 204)  # 204 No Content status

    file_type = scu.detect_file_type(demande.piece_jointe)
    suffix = ("." + file_type) if file_type not in ("unknown", "other_image") else ""
    return image_response(demande.piece_jointe, filename=f"{demande_num}{suffix}")


@bp.route("/demandes_inscriptions", defaults={"formsemestre_id": None})
@api_web_bp.route("/demandes_inscriptions", defaults={"formsemestre_id": None})
@bp.route("/formsemestre/<int:formsemestre_id>/demandes_inscriptions")
@api_web_bp.route("/formsemestre/<int:formsemestre_id>/demandes_inscriptions")
@login_required
@scodoc
@permission_required(Permission.DemandeInscription)
@as_json
def demandes_inscriptions(formsemestre_id: int | None = None):
    """
    Liste des demandes d'inscriptions.
    Accepte paramètres de pagination DataTables.
    Si include_all, inclue aussi les demandes annulées ou non finalisées (limbo)
    """
    include_all = request.args.get("include_all", type=int, default=False)
    only_attente = request.args.get("only_attente", type=int, default=False)
    formsemestre = (
        None
        if formsemestre_id is None
        else FormSemestre.get_formsemestre(formsemestre_id)
    )
    query, nb_total = FormSemestreDemandeInscription.build_query(
        formsemestre=formsemestre,
        only_attente=only_attente,
        include_all=include_all,
        search=request.args.get("search[value]"),
        scodoc_dept_id=g.scodoc_dept_id if g.scodoc_dept else None,
    )

    total_filtered = query.count()
    # Traitement de la pagination et tri pour DataTable server side
    # sorting
    order = []
    i = 0
    while True:
        col_index = request.args.get(f"order[{i}][column]")
        if col_index is None:
            break
        col_name = request.args.get(f"columns[{col_index}][data]")
        if col_name not in [
            "date_demande",
            "cours",
            "niveau",
            "statut",
            "civilite",
            "nom",
            "prenom",
            "email",
        ]:
            col_name = "date_demande"
        descending = request.args.get(f"order[{i}][dir]") == "desc"
        col = getattr(FormSemestreDemandeInscription, col_name)
        if descending:
            col = col.desc()
        order.append(col)
        i += 1
    if order:
        query = query.order_by(*order)
    # pagination
    start = request.args.get("start", type=int)
    length = request.args.get("length", type=int)
    query = query.offset(start).limit(length)
    # response
    return {
        "data": [demande.to_dict_table() for demande in query],
        "recordsFiltered": total_filtered,
        "recordsTotal": nb_total,
        "draw": request.args.get("draw", type=int),
    }
