# -*- coding: utf-8 -*-
##############################################################################
#
# Gestion scolarite IUT
#
# Copyright (c) 1999 - 2025 Emmanuel Viennet.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Emmanuel Viennet      emmanuel.viennet@viennet.net
#
##############################################################################

"""Vues AutoSco"""

from flask import render_template, url_for
from flask import flash, g, redirect, request

from app import db
from app.autosco import bp
from app.decorators import (
    scodoc,
    permission_required,
)
from app.forms.generic import SimpleConfirmationForm
from app.models import (
    FormSemestre,
    FormSemestreDemandeInscription,
)
from app.scodoc.sco_exceptions import ScoValueError
from app.scodoc.gen_tables import GenTable
from app.scodoc.sco_permissions import Permission
from app.scodoc import sco_preferences
from app.scodoc import sco_utils as scu
from app.views import ScoData
import sco_version


@bp.route("demandes_inscriptions", defaults={"formsemestre_id": -1})
@bp.route("demandes_inscriptions/formsemestre/<int:formsemestre_id>")
@scodoc
@permission_required(Permission.EtudInscrit)
def demandes_inscriptions(formsemestre_id: int):
    """Liste toutes les demandes d'inscriptions non annulées
    du département ou du formsemestre indiqué."""
    fmt = request.args.get("fmt", "html")
    if fmt not in ("html", "xls", "pdf"):
        raise ScoValueError("Format non supporté")
    include_all = int(request.args.get("include_all", False) == "1")
    only_attente = int(request.args.get("only_attente", False) == "1")
    formsemestre = (
        FormSemestre.get_formsemestre(formsemestre_id)
        if formsemestre_id != -1
        else None
    )

    if fmt == "html":
        return render_template(
            "autosco/formsemestre_demandes_inscriptions.j2",
            formsemestre=formsemestre,
            table_url=url_for(
                "apiweb.demandes_inscriptions",
                scodoc_dept=g.scodoc_dept,
            ),
            include_all=include_all,
            only_attente=only_attente,
            sco=ScoData(formsemestre=formsemestre),
            title=f"""Demandes d'inscriptions en {
                formsemestre.titre if formsemestre else g.scodoc_dept}""",
        )
    # Pour les formats xls et pdf, on construit la table
    query, _ = FormSemestreDemandeInscription.build_query(
        formsemestre=formsemestre,
        only_attente=only_attente,
        include_all=include_all,
        scodoc_dept_id=g.scodoc_dept_id if g.scodoc_dept else None,
    )
    table = _build_table_demandes(
        query.order_by(FormSemestreDemandeInscription.date_demande.desc()),
        preferences=sco_preferences.SemPreferences(),
        titre="Demandes d'inscriptions",
    )
    return table.make_page(fmt=fmt)


def _build_table_demandes(
    demandes: list[FormSemestreDemandeInscription],
    titre: str = "",
    preferences: sco_preferences.SemPreferences | None = None,
) -> GenTable:
    """Construit table des demandes d'inscriptions
    avec date_demande, cours, statut, niveau, civilite, nom, prénom, email
    et lien vers page de consultation de la demande.
    """
    rows = []
    for d in demandes:
        row = {
            "date_demande": d.date_demande.strftime(scu.DATEATIME_FMT),
            "_date_demande_xls": d.date_demande,
            "_date_demande_order": d.date_demande.isoformat(),
            "cours": d.formsemestre.titre,
            "statut": d.get_statut().version_lisible(),
            "niveau": d.niveau,
            "civilite": d.civilite,
            "nom": d.nom,
            "prenom": d.prenom,
            "email": d.email,
        }
        detail_url = url_for(
            "autosco.demande_inscription", numero=d.numero, scodoc_dept=g.scodoc_dept
        )
        # Ajoute lien dans chaque cas vers page détail:
        keys = list(row.keys())
        for key in keys:
            row[f"_{key}_target"] = detail_url
        rows.append(row)
    titles = {
        "date_demande": "Date de la demande",
        "cours": "Cours",
        "statut": "Statut",
        "niveau": "Niveau",
        "civilite": "Civilité",
        "nom": "Nom",
        "prenom": "Prénom",
        "email": "Email",
    }
    return GenTable(
        base_url=request.url,
        caption=f"Demandes d'inscriptions pour le semestre {titre}",
        columns_ids=(
            "date_demande",
            "cours",
            "statut",
            "niveau",
            "civilite",
            "nom",
            "prenom",
            "email",
        ),
        filename=scu.make_filename(f"demandes_inscription_{titre}"),
        html_class="table_demandes_inscriptions",
        html_sortable=True,
        origin=f"Généré par {sco_version.SCONAME} le {scu.timedate_human_repr()}",
        page_title=f"Demandes d'inscriptions pour le semestre {titre}",
        preferences=preferences,
        rows=rows,
        table_id="table_demandes_inscriptions",
        titles=titles,
    )


@bp.route("/demande_inscription/<string:numero>", methods=["GET"])
@scodoc
@permission_required(Permission.EtudInscrit)
def demande_inscription(numero: str):
    "Affiche une demande d'inscriptions pour ce semestre"
    demande: FormSemestreDemandeInscription = (
        db.session.query(FormSemestreDemandeInscription)
        .filter_by(numero=numero)
        .first_or_404()
    )
    formsemestre = FormSemestre.get_formsemestre(demande.formsemestre_id)

    image_piece_jointe = demande.piece_jointe and (
        scu.detect_file_type(demande.piece_jointe) in ("jpeg", "png")
    )

    # Si la demande a été acceptée, cherche l'inscription associée
    inscription = demande.get_inscription()

    return render_template(
        "autosco/demande_inscription.j2",
        demande=demande,
        formsemestre=formsemestre,
        nb_inscrits=len(formsemestre.get_inscrits(etats={scu.INSCRIT, scu.DEF})),
        image_piece_jointe=image_piece_jointe,
        inscription=inscription,
        sco=ScoData(formsemestre=formsemestre),
    )


@bp.route("/demande_inscription/accepte/<string:numero>", methods=["GET", "POST"])
@scodoc
@permission_required(Permission.EtudInscrit)
def demande_inscription_accepte(numero: str):
    "Accepte une demande d'inscriptions pour ce semestre"
    return _action_demande(numero, action="accepte")


@bp.route("/demande_inscription/refuse/<string:numero>", methods=["GET", "POST"])
@scodoc
@permission_required(Permission.EtudInscrit)
def demande_inscription_refuse(numero: str):
    "Accepte une demande d'inscriptions pour ce semestre"
    return _action_demande(numero, action="refuse")


def _action_demande(numero: str, action: str = "accept"):
    query = db.session.query(FormSemestreDemandeInscription).filter_by(numero=numero)
    if g.scodoc_dept:
        query = query.join(FormSemestre).filter_by(dept_id=g.scodoc_dept_id)
    demande: FormSemestreDemandeInscription = query.first_or_404()
    if not demande.modifiable():
        raise ScoValueError(f"Demande non modifiable (statut: {demande.get_statut()})")
    form = SimpleConfirmationForm()
    if form.validate_on_submit():
        match action:
            case "refuse":
                demande.refuse()
                flash("demande refusée")
            case "accepte":
                demande.accept()
                flash("demande acceptée")
            case _:
                raise ScoValueError("Action non supportée")

        return redirect(
            url_for(
                "autosco.demande_inscription",
                scodoc_dept=demande.formsemestre.departement.acronym,
                numero=demande.numero,
            )
        )

    return render_template(
        "autosco/demande_inscription_action_confirm.j2",
        action=action,
        demande=demande,
        formsemestre=demande.formsemestre,
        form=form,
        sco=ScoData(formsemestre=demande.formsemestre),
    )
