"""ScoDoc models : templates pour documents docx.

template is a binary blob, docx document with jinja markup.
"""

from flask import g, url_for
from flask_login import current_user

from app import db, models
from app.scodoc.sco_exceptions import ScoException, ScoPermissionDenied, ScoValueError
from app.scodoc.sco_permissions import Permission


class DocxTemplate(models.ScoDocModel):
    """Template docx pour les documents générés par ScoDoc"""

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Text(), nullable=False, index=True)
    description = db.Column(db.Text(), nullable=False, default="", server_default="")
    template = db.Column(db.LargeBinary(), nullable=False)
    """docx template, with jinja markup"""
    template_type = db.Column(
        db.Enum("lettre_individuelle", "pv_jury", "autre", name="docx_template_type"),
        nullable=False,
        index=True,
    )
    """Type de document : pv jury, lettre individuelle, etc.
    """
    dept_id = db.Column(
        db.Integer, db.ForeignKey("departement.id", ondelete="CASCADE"), index=True
    )
    """Département auquel appartient le document.
    Si None, template accessible dans tous les départements.
    """
    author_id = db.Column(db.Integer, db.ForeignKey("user.id", ondelete="SET NULL"))
    author = db.relationship("User", back_populates="docx_templates")
    """Utilisateur qui a créé le document.
    """
    date_creation = db.Column(
        db.DateTime(timezone=True), server_default=db.func.now(), nullable=False
    )
    """Date de création du document.
    """
    date_modification = db.Column(
        db.DateTime(timezone=True), server_default=db.func.now(), onupdate=db.func.now()
    )
    """Date de dernière modification du document.
    """

    def __str__(self):
        return f"""DocxTemplate(name='{self.name}', type={self.template_type
        }, author={self.author}, [{len(self.template)} bytes])"""

    def can_be_edited_by(self, user=None) -> bool:
        """Vérifie si le template peut être modifié ou supprimé par l'utilisateur.
        Pour supprimer le template :
        - si attaché à un dept: il faut avoir le droit EditPreferences dans ce dept
        - si pas attaché à un dept: il faut être admin
        """
        user = user or current_user
        if self.dept_id:
            return user.has_permission(Permission.EditPreferences, dept_id=self.dept_id)
        return user.is_administrator()

    def can_be_viewed_by(self, user=None) -> bool:
        """Vérifie si le template peut être vu par l'utilisateur (default: current).
        Pour voir le template :
        - si attaché à un dept: il faut avoir le droit View dans ce dept
        - si pas attaché à un dept: tout le monde
        """
        # tout le monde peut voir les templates globaux,
        # mais pas ceux des départements sur lesquels on n'a pas le droit
        user = user or current_user
        return (not self.departement) or current_user.has_permission(
            Permission.ScoView, self.departement.acronym
        )

    @classmethod
    def get_dept_templates(
        cls, dept_id=None, template_type: str = ""
    ) -> tuple[list["DocxTemplate"], list["DocxTemplate"]]:
        """Returns list of global templates and dept templates of the given type"""
        global_templates = (
            DocxTemplate.query.filter_by(dept_id=None, template_type=template_type)
            .order_by(DocxTemplate.name)
            .all()
        )
        dept_templates = (
            (
                DocxTemplate.query.filter_by(
                    dept_id=dept_id, template_type=template_type
                )
                .order_by(DocxTemplate.name)
                .all()
            )
            if dept_id
            else []
        )
        return global_templates, dept_templates

    @classmethod
    def get_template_docx(cls, template_id: int, template_type: str = "") -> bytes:
        """Renvoie le document docx du template d'id donné.
        Si id==-1, renvoie le document par défaut par défaut pour le type indiqué
        """
        if template_id == -1:
            # use default ScoDoc template
            if template_type not in DocxTemplate.template_type.type.enums:
                raise ScoException("Template type non valide")
            filename = f"/opt/scodoc/app/templates/docx/template_{template_type}.docx"
            try:
                with open(filename, "rb") as f:
                    return f.read()
            except FileNotFoundError as exc:
                raise ScoException(
                    f"Template par défaut non trouvé : {filename}"
                ) from exc
        template = cls.query.get(template_id)
        if not template:
            raise ScoValueError("Template non trouvé")
        if not template.can_be_viewed_by():
            raise ScoPermissionDenied(
                dest_url=url_for("scolar.index", scodoc_dept=g.scodoc_dept)
            )
        return template.template

    @classmethod
    def get_template_type_str(cls, template_type: str) -> str:
        "Version française du type de template (au pluriel)"
        return {
            "lettre_individuelle": "lettres individuelles de PV de jury",
            "pv_jury": "procès-verbaux (PV) de jury",
            "autre": "autres documents",
        }.get(template_type, "inconnu")

    def template_type_str(self) -> str:
        "Version française du type de template (au pluriel)"
        return self.__class__.get_template_type_str(self.template_type)
