##############################################################################
# ScoDoc
# Copyright (c) 1999 - 2024 Emmanuel Viennet.  All rights reserved.
# See LICENSE
##############################################################################

"""Demandes d'inscription à un formsemestre pour applications tierces.

Les demandes d'inscriptions peuvent être créées via l'API uniquement.
Dans ScoDoc, elles sont consultables et on peut les valider.
"""

import datetime
import re
import secrets
import string
import time

from flask import render_template
import sqlalchemy as sa

from app import db, log, models
from app import email
from app.models import (
    Adresse,
    APO_CODE_STR_LEN,
    FormSemestre,
    FormSemestreInscription,
    Identite,
)
from app.models.config import ScoDocSiteConfig
from app.models.etudiants import input_civilite
from app.scodoc.sco_exceptions import ScoValueError
from app.scodoc import sco_etud, sco_preferences
from app.scodoc import sco_utils as scu


class StatutDemandeInscription(int, scu.BiDirectionalEnum):
    """Valeurs pour le champ statut des FormSemestreDemandeInscription"""

    # Stockés en BD ne pas modifier:
    INVALID = -1
    "état invalide"
    LIMBO = 0
    "demande déposée, mais mail non vérifié"
    ATTENTE = 1
    "demande déposée, en attente de validation"
    ACCEPTEE = 2
    "demande acceptée, étudiant inscrit"
    REFUSEE = 3
    "demande refusée, étudiant non inscrit"
    ANNULEE = 4
    "demande annulée (avant validation)"

    def version_lisible(self) -> str:
        """Retourne une version lisible de l'état de la demande"""
        return {
            StatutDemandeInscription.INVALID: "invalide",
            StatutDemandeInscription.LIMBO: "dépôt non finalisé",
            StatutDemandeInscription.ATTENTE: "en attente",
            StatutDemandeInscription.ACCEPTEE: "acceptée",
            StatutDemandeInscription.REFUSEE: "refusée",
            StatutDemandeInscription.ANNULEE: "annulée",
        }.get(self, "")

    @classmethod
    def is_valid_statut(cls, etat: int) -> bool:
        "True if etat is valid"
        return (
            etat in cls._value2member_map_
            and etat != StatutDemandeInscription.INVALID.value
        )


def base36_encode(number):
    """Encodes a number into base36 (0-9, A-Z)"""
    chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    if number == 0:
        return "0"
    result = ""
    while number:
        number, i = divmod(number, 36)
        result = chars[i] + result
    return result


def generate_timestamped_id(groups: int = 1, group_size: int = 5) -> str:
    """Generate a timestamped ID with random alphanumeric parts."""
    # Attention : si modification, modifier aussi le code du client
    # (pour AutoSco: utils.validate_numero)
    chars = string.ascii_uppercase + string.digits

    # Get current time as integer (e.g., milliseconds)
    timestamp = int(time.time() * 1000)
    encoded_ts = base36_encode(timestamp)  # Short base36 timestamp

    random_part = "-".join(
        "".join(secrets.choice(chars) for _ in range(group_size)) for _ in range(groups)
    )

    return f"{encoded_ts}-{random_part}"


def datetime_from_id(timestamped_id) -> datetime.datetime:
    """Convert a timestamped ID back to a datetime object.
    Used for debugging.
    """
    # Extract the first part (timestamp is before the first dash)
    encoded_ts = timestamped_id.split("-")[0]
    ts_ms = int(encoded_ts, 36)  # milliseconds since epoch
    ts_seconds = ts_ms / 1000
    return datetime.datetime.fromtimestamp(ts_seconds)


class FormSemestreDemandeInscription(models.ScoDocModel):
    """Informations décrivant un "semestre" (session) de formation
    pour un apprenant.
    """

    _sco_dept_relations = ["formsemestre"]

    id = db.Column(db.Integer, primary_key=True)
    formsemestre_id = db.Column(
        db.Integer,
        db.ForeignKey("notes_formsemestre.id", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    formsemestre = db.relationship(
        "FormSemestre", back_populates="demandes_inscriptions"
    )
    date_demande = db.Column(
        db.DateTime(timezone=True),
        default=datetime.datetime.now,
        nullable=False,
        server_default=db.func.now(),
    )
    "date de la demande"
    date_validation = db.Column(db.DateTime(timezone=True), nullable=True)
    "date de validation"
    statut = db.Column(
        db.Integer,
        nullable=False,
        default=StatutDemandeInscription.LIMBO.value,
        server_default=str(StatutDemandeInscription.LIMBO.value),
    )
    "voir StatutDemandeInscription"
    numero = db.Column(
        db.Text(),
        nullable=False,
        default="",
        server_default="",
        unique=True,
        index=True,
    )
    "numéro de la demande (uuid), généré lors du dépot de la demande"
    # --- Informations sur l'apprenant
    # Codes INE et NIP non requis car inconnus pour les apprenants extérieurs
    code_nip = db.Column(db.Text(), nullable=False, default="", server_default="")
    code_ine = db.Column(db.Text(), nullable=False, default="", server_default="")
    etape = db.Column(
        db.String(APO_CODE_STR_LEN), nullable=False, default="", server_default=""
    )
    "optionnel, étape principale d'inscription pour les étudiants locaux"
    civilite = db.Column(db.String(1), nullable=False)
    "civilité de l'apprenant (M, F, X)"
    nom = db.Column(db.Text(), nullable=False, default="", server_default="")
    "nom de l'apprenant"
    prenom = db.Column(db.Text(), nullable=False, default="", server_default="")
    "prénom de l'apprenant"
    email = db.Column(
        db.Text(), nullable=False, default="", server_default="", index=True
    )
    "email de l'apprenant"
    telephone = db.Column(db.Text(), nullable=False, default="", server_default="")
    "téléphone de l'apprenant"
    niveau = db.Column(db.Text(), nullable=False, default="", server_default="")
    "niveau de l'apprenant (texte libre, normalement 'debutant', 'intermediaire', 'avance')"
    motivation = db.Column(db.Text(), nullable=False, default="", server_default="")
    "motivation de l'apprenant (texte libre)"
    piece_jointe = db.Column(db.LargeBinary(), nullable=True)
    "pièce jointe (pdf, image, etc.)"

    __table_args__ = (
        db.CheckConstraint("civilite IN ('M', 'F', 'X')"),  # non nullable
    )

    @classmethod
    def generate_numero(cls) -> str:
        "renvoie nouveau numéro (chaîne alphanumérique), non existant en base."
        numero = generate_timestamped_id()
        while cls.query.filter_by(numero=numero).first() is not None:
            numero = str(generate_timestamped_id())
        return numero

    @classmethod
    def create(cls, **kwargs) -> "FormSemestreDemandeInscription":
        """Crée une demande d'inscription. Vérifie les arguments.
        Le statut est mis à ATTENTE
        """
        if not ScoDocSiteConfig.is_autosco_enabled():
            raise ScoValueError("autosco non activé")
        kwargs = cls.convert_dict_fields(kwargs)
        errs = cls.check_args(kwargs)
        if errs:
            raise ScoValueError(errs)
        numero = cls.generate_numero()
        date_demande = datetime.datetime.now()
        obj = cls(
            date_demande=date_demande,
            numero=numero,
            statut=StatutDemandeInscription.ATTENTE.value,
            **kwargs,
        )
        log(f"FormSemestreDemandeInscription: created {obj!r}")
        return obj

    def delete(self):
        "Supprime la demande d'inscription, log (do not commit)"
        log(f"FormSemestreDemandeInscription: deleting {self!r}")
        db.session.delete(self)

    def __repr__(self):
        return f"""<FormSemestreDemandeInscription {self.numero} ({self.nom}) {
            self.formsemestre} {self.get_statut().name}>"""

    def from_dict(
        self, args: dict, excluded: set[str] | None = None, log: bool = False
    ) -> list[str]:
        """Update object's fields given in dict. Add to session but don't commit.
        If log, write to ScoDoc log.
        Returns list of modified fields.
        """
        errs = self.check_args(args, require_args=False)
        if errs:
            raise ScoValueError(f"invalid args: {errs}")
        super().from_dict(args, excluded=excluded, log=log)

    def to_dict(self, exclude_images=True) -> dict:
        "dict, tous les attributs sauf les images (pièce jointe)"
        d = dict(self.__dict__)
        d.pop("_sa_instance_state", None)
        d.pop("formsemestre", None)
        d["statut"] = self.get_statut().name

        if exclude_images:
            d.pop("piece_jointe", None)
        return d

    def to_dict_table(self) -> dict:
        "dict, avec attributs pour table HTML"
        d = self.to_dict(exclude_images=True)
        d["cours"] = self.formsemestre.titre
        return d

    @classmethod
    def convert_dict_fields(cls, args: dict) -> dict:
        """Convert fields from the given dict to model's attributes values.
        Converti statut chaîne en int.
        Remove keys not in model.
        args est copié et retourné.
        """
        args = {k: v for k, v in args.items() if k in cls.__table__.columns}
        super().convert_dict_fields(args)
        if "statut" in args:
            if isinstance(args["statut"], StatutDemandeInscription):
                args["statut"] = args["statut"].value
            elif isinstance(args["statut"], int):
                try:
                    args["statut"] = StatutDemandeInscription(args["statut"]).value
                except ValueError:
                    args["statut"] = StatutDemandeInscription.INVALID.value
            elif isinstance(args["statut"], str):
                statut = StatutDemandeInscription.get(args["statut"])
                if statut is None:
                    raise ScoValueError("statut invalide")
                args["statut"] = statut.value
        return args

    @classmethod
    def check_args(
        cls,
        kwargs: dict,
        allow_statut: bool = False,
        require_args: bool = True,
    ) -> str:
        """Check that dict contains valid arguments to create a new instance.
        If require_args, check that all required args are present.
        If allow_statut, allow to modify the statut.
        Returns empty string if ok, else error message.
        Certains messages sont en français pour affichage sur le client API.
        """
        if "id" in kwargs:
            return "can't specify id"
        if require_args:
            formsemestre_id = kwargs.get("formsemestre_id", None)
            if formsemestre_id is None:
                return "missing required arg: formsemestre_id"
            formsemestre = FormSemestre.get_formsemestre(
                formsemestre_id, accept_none=True
            )
            if formsemestre is None:
                return "invalid formsemestre_id"
        if "date_demande" in kwargs:
            return "can't specify date_demande"
        if "date_validation" in kwargs:
            return "can't specify date_validation"
        if "statut" in kwargs and not allow_statut:
            return "can't specify statut"
        if "numero" in kwargs:
            return "can't specify numero"
        # Check required fields
        for field in ("civilite", "nom", "prenom", "motivation"):
            if field in kwargs:
                if not isinstance(kwargs[field], str):
                    return f"invalid type for {field}"
                kwargs[field] = kwargs[field].strip()
                if not kwargs[field]:
                    return f"champs {field} requis"  # en français, sera affiché à l'utilisateur
            elif require_args:
                return f"champs {field} requis"  # en français, sera affiché à l'utilisateur
        # Check civilite
        if "civilite" in kwargs:
            kwargs["civilite"] = input_civilite(kwargs["civilite"])
        # Check email regexp
        if "email" in kwargs:
            if not isinstance(kwargs["email"], str):
                return "invalid type for email"
            kwargs["email"] = kwargs["email"].strip()
            email_addr = kwargs["email"]
            if not scu.is_valid_mail(email_addr):
                return "Adresse email invalide"
        elif require_args:
            return "champ email requis"
        # Check tel regexp
        if "telephone" in kwargs:
            if not isinstance(kwargs["telephone"], str):
                return "invalid type for telephone"
            kwargs["telephone"] = kwargs["telephone"].strip()
            telephone = kwargs["telephone"]
            if not re.match(r"^\+?[0-9\ \.]+$", telephone):
                return "Numéro de téléphone invalide"
        elif require_args:
            return "champ téléphone requis"
        # Check niveau
        if "niveau" in kwargs:
            if not isinstance(kwargs["niveau"], str):
                return "invalid type for niveau"
            kwargs["niveau"] = kwargs["niveau"].strip().lower()
            if kwargs["niveau"] not in ("debutant", "intermediaire", "avance"):
                return "champ 'niveau' invalide (debutant, intermediaire, avance)"
        elif require_args:
            return "champ niveau requis"
        # Check piece jointe
        kwargs["piece_jointe"] = kwargs.get("piece_jointe", None)
        if kwargs["piece_jointe"] is not None and not isinstance(
            kwargs["piece_jointe"], bytes
        ):
            return "invalid piece_jointe"
        if kwargs["piece_jointe"] is not None:
            detect_file_type_result = scu.detect_file_type(kwargs["piece_jointe"])
            if detect_file_type_result not in ("jpeg", "png", "pdf"):
                return "Format de la pièce jointe invalide (JPEG, PNG ou PDF requis)"
        return ""

    def get_statut(self) -> StatutDemandeInscription:
        "le statut de la demande (une Enum)"
        try:
            return StatutDemandeInscription(self.statut)
        except ValueError:
            return StatutDemandeInscription.INVALID

    def modifiable(self) -> bool:
        """Vérifie si la demande d'inscription est modifiable.
        Une demande est modifiable si elle est en attente ou refusée
        """
        statut = self.get_statut()
        return statut in (
            StatutDemandeInscription.ATTENTE,
            StatutDemandeInscription.REFUSEE,
        )

    def describe(self) -> str:
        "Renvoie une description de la demande d'inscription"
        formsemestre: FormSemestre = self.formsemestre
        if not formsemestre:
            return "demande invalide (pas de semestre)"
        return (
            f"Nom: {self.nom}\n"
            f"Prénom: {self.prenom}\n"
            f"Email: {self.email}\n"
            f"Téléphone: {self.telephone}\n\n"
            f"Demande d'inscription à {formsemestre.titre_formation()}\n"
            f"Niveau: {self.niveau}\n"
            f"Date de la demande: {self.date_demande.strftime('%Y-%m-%d %H:%M:%S')}\n"
            f"Statut: {self.get_statut().version_lisible()}\n"
            f"Numéro de la demande: {self.numero}\n"
        )

    def _mail_etud(self, subject: str, template_name: str = "", body: str = ""):
        """envoi d'un mail au candidat.
        Si template_name, utilise le template indiqué, sinon fixed_content est utilisé.
        """
        formsemestre: FormSemestre = self.formsemestre
        if not formsemestre:
            return
        prefs = sco_preferences.SemPreferences(
            formsemestre.id, dept_id=formsemestre.dept_id
        )
        subject = f"""[{prefs['autosco_organisation_name'] or 'ScoDoc'
            }] {subject}"""
        if template_name:
            # Si template_name, utilise le template standard à la place de body
            body = render_template(
                template_name,
                demande=self,
                formsemestre=formsemestre,
                prefs=prefs,
            )
        if body:
            sender = prefs["autosco_from_addr"] or email.get_from_addr()
            email.send_email(subject, sender, [self.email], body)
        else:
            log(f"FormSemestreDemandeInscription: pas de corps de mail pour {self!r}")

    def _mail_responsable(self, subject: str, template_name: str = ""):
        "envoi d'un mail au responsable AutoSco"
        formsemestre: FormSemestre = self.formsemestre
        if not formsemestre:
            return
        prefs = sco_preferences.SemPreferences(
            formsemestre.id, dept_id=formsemestre.dept_id
        )
        if not prefs["autosco_responsable_addr"]:
            log(
                f"FormSemestreDemandeInscription: pas d'adresse responsable pour {formsemestre!r}"
            )
            return
        subject = f"""[{prefs['autosco_organisation_name'] or 'ScoDoc'
            }] {subject}"""
        body = render_template(
            template_name,
            demande=self,
            formsemestre=formsemestre,
            prefs=prefs,
        )
        sender = prefs["autosco_from_addr"] or email.get_from_addr()
        email.send_email(subject, sender, [prefs["autosco_responsable_addr"]], body)

    def notify_demande(self):
        """Envoie un mail à l'apprenant pour lui dire que sa demande a été enregistrée, et notifie le responsable AutoSco."""
        log(f"FormSemestreDemandeInscription: notify demande {self!r}")
        formsemestre: FormSemestre = self.formsemestre
        if not formsemestre:
            return
        self._mail_etud(
            subject=f"""Demande d'inscription à {formsemestre.titre} enregistrée""",
            template_name="autosco/mail_ack_demande.txt",
        )
        self._mail_responsable(
            subject=f"""Nouvelle demande d'inscription à {formsemestre.titre}""",
            template_name="autosco/mail_notif_resp_demande.txt",
        )

    def notify_acceptation(self):
        """Notification après acceptation de la demande"""
        log(f"FormSemestreDemandeInscription: notify acceptation {self!r}")

        prefs = sco_preferences.SemPreferences(
            self.formsemestre_id, dept_id=self.formsemestre.dept_id
        )
        body = render_template(
            "autosco/mail_accept_demande.txt",
            demande=self,
            formsemestre=self.formsemestre,
            prefs=prefs,
            information_nouvel_inscrit=prefs["autosco_mail_information_nouvel_inscrit"],
        )

        self._mail_etud(
            subject=f"""Confirmation d'inscription à {self.formsemestre.titre}""",
            body=body,
        )

    def notify_refus(self):
        """Notification après refus de la demande"""
        log(f"FormSemestreDemandeInscription: notify refus {self!r}")
        prefs = sco_preferences.SemPreferences(
            self.formsemestre_id, dept_id=self.formsemestre.dept_id
        )
        body = (prefs["autosco_mail_refus_tmpl"] or "").strip()

        self._mail_etud(
            subject=f"""Refus d'inscription à {self.formsemestre.titre}""",
            template_name="autosco/mail_refus_demande.txt" if not body else "",
            body=body,
        )

    def process_auto(self):
        """Traitement automatique de la demande.
        Appelée au moment du dépot de la demande, après sa création.
        Suivant le réglage de autosco_accept_mode, accepte automatiquement
        ou non la demande. Si accepte, l'étudiant est notifié (méthode accept).
        """
        mode = sco_preferences.get_preference(
            "autosco_accept_mode", self.formsemestre_id
        )
        is_local = self.code_ine or self.code_nip
        if (mode == "all") or ((mode == "local") and is_local):
            if not self.formsemestre.capacite_accueil_atteinte():
                self.accept()

    def accept(self) -> Identite:
        """Accepte la demande, si capacité d'acceuil le permet:
        inscrit l'étudiant au formsemestre et le notifie.
        """
        from app.scodoc import sco_formsemestre_inscriptions

        if self.formsemestre.capacite_accueil_atteinte():
            raise ScoValueError("Capacité d'accueil atteinte, demande non acceptée")
        if self.statut == StatutDemandeInscription.ACCEPTEE:
            raise ScoValueError("Demande déjà acceptée.")
        etud = self.get_or_create_etud()
        # Inscrit au formsemestre
        sco_formsemestre_inscriptions.do_formsemestre_inscription_with_modules(
            self.formsemestre_id,
            etud.id,
            dept_id=self.formsemestre.dept_id,
            etape=self.etape,
            method="autosco.accept",
        )
        log(f"FormSemestreDemandeInscription: accepte {self!r}")
        log(f"formsemestre_etud_inscrit: {etud} inscrit à {self.formsemestre}")
        # Change le statut
        self.statut = StatutDemandeInscription.ACCEPTEE.value
        self.date_validation = datetime.datetime.now()
        db.session.add(self)
        db.session.commit()
        db.session.refresh(self)
        # Envoie un mail à l'étudiant pour lui dire qu'il est inscrit
        self.notify_acceptation()
        return etud

    def refuse(self):
        """Refuse la demande d'inscription.
        Change le statut et envoie un mail à l'étudiant.
        """
        if self.statut == StatutDemandeInscription.REFUSEE:
            raise ScoValueError("Demande déjà refusée.")
        self.statut = StatutDemandeInscription.REFUSEE.value
        self.date_validation = None
        db.session.add(self)
        db.session.commit()
        db.session.refresh(self)
        log(f"FormSemestreDemandeInscription: refuse {self!r}")
        self.notify_refus()

    def get_or_create_etud(self) -> Identite:
        """Récupère l'étudiant correspondant à la demande d'inscription.
        Crée l'étudiant s'il n'existe pas.
        """
        etud = Identite.query.join(Adresse).filter_by(email=self.email).first()
        if etud is None:
            # Aucun étudiant ScoDoc avec ce mail: on doit le créer
            ok, _ = sco_etud.check_nom_prenom_homonyms(nom=self.nom, prenom=self.prenom)
            if not ok:
                raise ScoValueError("nom ou prénom invalide")
            etud = Identite.create_etud(
                civilite=self.civilite,
                nom=self.nom,
                prenom=self.prenom,
                code_ine=self.code_ine,
                code_nip=self.code_nip,
            )
            db.session.commit()
            db.session.refresh(etud)
            # Set address
            adr = etud.adresses.first()
            adr.email = self.email
            adr.telephone = self.telephone
            db.session.add(adr)
            db.session.commit()
            db.session.refresh(etud)
        return etud

    @classmethod
    def build_query(
        cls,
        formsemestre: FormSemestre | None = None,
        scodoc_dept_id: int | None = None,
        include_all: int | bool = False,
        only_attente: int | bool = False,
        search: str = "",
    ) -> tuple:
        """Construit la requête pour récupérer les demandes d'inscriptions.
        Si include_all, inclut les demandes annulées.
        Si only_attente, filtre pour ne garder que les demandes en attente.
        Si search, filtre les demandes par chaîne de caractères
        Renvoie: (query, total)
        où query est une requête SQLAlchemy filtrée et total est le
        nombre total de demandes du département et formsemestre (si spécifié).
        """
        query = (
            formsemestre.demandes_inscriptions
            if formsemestre
            else FormSemestreDemandeInscription.query
        )
        query = query.join(FormSemestre)
        if scodoc_dept_id is not None:
            # Si route départementale, élimine les demandes d'autres départements
            query = query.filter(FormSemestre.dept_id == scodoc_dept_id)
        nb_total = query.count()

        if only_attente:
            query = query.filter(
                FormSemestreDemandeInscription.statut
                == StatutDemandeInscription.ATTENTE
            )
        elif not include_all:
            # Filtre les demandes annulées ou non finalisées (limbo)
            query = query.filter(
                db.and_(
                    FormSemestreDemandeInscription.statut
                    != StatutDemandeInscription.ANNULEE,
                    FormSemestreDemandeInscription.statut
                    != StatutDemandeInscription.LIMBO,
                )
            )
        # Recherche par chaine de caractères
        if search:
            query = query.filter(
                db.or_(
                    FormSemestreDemandeInscription.numero.ilike(f"%{search}%"),
                    FormSemestreDemandeInscription.code_nip.ilike(f"%{search}%"),
                    FormSemestreDemandeInscription.code_ine.ilike(f"%{search}%"),
                    FormSemestreDemandeInscription.nom.ilike(f"%{search}%"),
                    FormSemestreDemandeInscription.prenom.ilike(f"%{search}%"),
                    FormSemestreDemandeInscription.email.ilike(f"%{search}%"),
                    # cherche aussi dans titre cours:
                    FormSemestre.titre.ilike(f"%{search}%"),
                )
            )
        return query, nb_total

    def get_inscription(self) -> None | FormSemestreInscription:
        """Retourne l'inscription associée à la demande d'inscription."""
        if (
            self.get_statut() != StatutDemandeInscription.ACCEPTEE
            or not self.formsemestre
        ):
            # Si la demande n'est pas acceptée, on ne peut pas chercher l'inscription
            return None
        # Si la demande a été acceptée, cherche l'inscription associée
        stmt = (
            sa.select(FormSemestreInscription)
            .filter_by(formsemestre_id=self.formsemestre.id)
            .join(Identite)
        )
        if self.code_nip:
            stmt = stmt.filter_by(code_nip=self.code_nip)
        elif self.code_ine:
            stmt = stmt.filter_by(code_ine=self.code_ine)
        elif self.email:
            # Si pas de code NIP ou INE, on cherche par email
            stmt = stmt.join(Adresse).filter(Adresse.email == self.email)
        return db.session.execute(stmt).scalars().first()
