# -*- mode: python -*-
# -*- coding: utf-8 -*-

##############################################################################
#
# Gestion scolarite IUT
#
# Copyright (c) 1999 - 2025 Emmanuel Viennet.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Emmanuel Viennet      emmanuel.viennet@viennet.net
#
##############################################################################

"""Edition des lettres individuelles de jury"""
# code initialement dans sco_pvpdf.py

import io
import re

from PIL import Image as PILImage
from PIL import UnidentifiedImageError

import reportlab
from reportlab.lib.units import cm, mm
from reportlab.lib.enums import TA_LEFT
from reportlab.platypus import PageBreak, Table, Image
from reportlab.platypus.doctemplate import BaseDocTemplate
from reportlab.lib import styles

from app.models import FormSemestre, Identite, ValidationDUT120

import app.scodoc.sco_utils as scu
from app.scodoc import sco_bulletins
from app.scodoc import sco_bulletins_pdf
from app.scodoc import sco_pv_dict
from app.scodoc import sco_pdf
from app.scodoc import sco_preferences
from app.scodoc.sco_exceptions import ScoPDFFormatError, ScoValueError
from app.scodoc.sco_cursus_dut import SituationEtudCursus
from app.scodoc.sco_pv_templates import CourrierIndividuelTemplate, jury_titres
import sco_version


def filter_decisions(decisions: list[dict]) -> list[dict]:
    """Renvoie la liste des décisions avec décision de jury"""
    return [
        decision
        for decision in decisions
        if (
            decision["decision_sem"]
            or decision.get("decision_annee")
            or decision.get("decision_rcue")
            or decision.get("decisions_ue")
        )
    ]


def get_lettres_param(
    formsemestre: FormSemestre,
    date_jury="",
    date_commission="",
    signature=None,
):
    """Renvoie un dictionnaire de paramètres pour les lettres individuelles
    (clés communes à toutes les lettres)
    """
    params = {
        "date_jury": date_jury,
        "date_commission": date_commission,
        "titre_formation": formsemestre.formation.titre_officiel,
        "htab1": "8cm",  # lignes à droite (entete, signature)
        "htab2": "1cm",
    }
    # copie preferences
    for name in sco_preferences.get_base_preferences().prefs_name:
        params[name] = sco_preferences.get_preference(name, formsemestre.id)
    return params


def pdf_lettres_individuelles(
    formsemestre_id: int,
    etudids=None,
    date_jury="",
    date_commission="",
    signature=None,
) -> bytes:
    """Document PDF avec les lettres d'avis pour les etudiants mentionnés
    (tous ceux du semestre, ou la liste indiquée par etudids)
    Renvoie pdf data ou vide si aucun etudiant avec décision de jury.
    """
    dpv = sco_pv_dict.dict_pvjury(formsemestre_id, etudids=etudids, with_prev=True)
    if not dpv:
        return b""
    #
    formsemestre = FormSemestre.get_formsemestre(formsemestre_id)
    params = get_lettres_param(
        formsemestre, date_jury=date_jury, date_commission=date_commission
    )
    decisions = filter_decisions(dpv["decisions"])
    if not decisions:
        return b""

    bookmarks = {}
    objects = []  # list of PLATYPUS objects
    for doc_number, decision in enumerate(decisions, start=1):
        etud = Identite.get_etud(decision["identite"]["etudid"])
        params["nomEtud"] = etud.nomprenom  # backward compatibility
        params.update(dict_pv_lettre_individuelle(formsemestre, etud, decision))

        bookmarks[doc_number] = scu.suppress_accents(etud.nomprenom)
        try:
            objects += pdf_lettre_individuelle(
                formsemestre, decision, params, signature
            )
        except UnidentifiedImageError as exc:
            raise ScoValueError(
                "Fichier image (signature ou logo ?) invalide !"
            ) from exc
        objects.append(PageBreak())

    # Paramètres de mise en page
    prefs = sco_preferences.SemPreferences(formsemestre.id)
    margins = (
        prefs["left_margin"],
        prefs["top_margin"],
        prefs["right_margin"],
        prefs["bottom_margin"],
    )

    # ----- Build PDF
    report = io.BytesIO()  # in-memory document, no disk file
    document = BaseDocTemplate(report)
    document.addPageTemplates(
        CourrierIndividuelTemplate(
            document,
            author=f"{sco_version.SCONAME} {sco_version.SCOVERSION} (E. Viennet)",
            title=f"Lettres décision {formsemestre.titre_annee()}",
            subject="Décision jury",
            margins=margins,
            pagesbookmarks=bookmarks,
            preferences=prefs,
        )
    )

    try:
        document.build(objects)
    except (ValueError, KeyError, reportlab.platypus.doctemplate.LayoutError) as exc:
        raise ScoPDFFormatError(str(exc)) from exc

    data = report.getvalue()
    return data


def _simulate_br(paragraph_txt: str, para="<para>") -> str:
    """Reportlab bug turnaround (could be removed in a future version).
    p is a string with Reportlab intra-paragraph XML tags.
    Replaces <br> (currently ignored by Reportlab) by </para><para>
    Also replaces <br> by <br/>
    """
    return ("</para>" + para).join(
        re.split(r"<.*?br.*?/>", paragraph_txt.replace("<br>", "<br/>"))
    )


def _make_signature_image(signature, leftindent, formsemestre_id) -> Table:
    "crée un paragraphe avec l'image signature"
    # cree une image PIL pour avoir la taille (W,H)

    f = io.BytesIO(signature)
    img = PILImage.open(f)
    width, height = img.size
    pdfheight = (
        1.0
        * sco_preferences.get_preference("pv_sig_image_height", formsemestre_id)
        * mm
    )
    f.seek(0, 0)

    style = styles.ParagraphStyle({})
    style.leading = 1.0 * sco_preferences.get_preference(
        "SCOLAR_FONT_SIZE", formsemestre_id
    )  # vertical space
    style.leftIndent = leftindent
    return Table(
        [("", Image(f, width=width * pdfheight / float(height), height=pdfheight))],
        colWidths=(9 * cm, 7 * cm),
    )


def pdf_lettre_individuelle(
    formsemestre: FormSemestre,
    decision: dict,
    params: dict,
    signature=None,
):
    """
    Renvoie une liste d'objets PLATYPUS pour intégration
    dans un autre document.
    """
    objects = []
    style = reportlab.lib.styles.ParagraphStyle({})
    style.fontSize = 14
    style.fontName = sco_preferences.get_preference("PV_FONTNAME", formsemestre.id)
    style.leading = 18
    style.alignment = TA_LEFT

    # Corps de la lettre:
    objects += sco_bulletins_pdf.process_field(
        sco_preferences.get_preference("PV_LETTER_TEMPLATE", formsemestre.id),
        params,
        style,
        suppress_empty_pars=True,
    )

    # Signature:
    # nota: si semestre terminal, signature par directeur IUT, sinon, signature par
    # chef de département.
    if decision["Se"].semestre_non_terminal:
        sig = (
            sco_preferences.get_preference(
                "PV_LETTER_PASSAGE_SIGNATURE", formsemestre.id
            )
            or ""
        ) % params
        sig = _simulate_br(sig, '<para leftindent="%(htab1)s">')
        objects += sco_pdf.make_paras(
            (
                """<para leftindent="%(htab1)s" spaceBefore="25mm">"""
                + sig
                + """</para>"""
            )
            % params,
            style,
        )
    else:
        sig = (
            sco_preferences.get_preference(
                "PV_LETTER_DIPLOMA_SIGNATURE", formsemestre.id
            )
            or ""
        ) % params
        sig = _simulate_br(sig, '<para leftindent="%(htab1)s">')
        objects += sco_pdf.make_paras(
            (
                """<para leftindent="%(htab1)s" spaceBefore="8mm">"""
                + sig
                + """</para>"""
            )
            % params,
            style,
        )

    if signature:
        try:
            objects.append(
                _make_signature_image(signature, params["htab1"], formsemestre.id)
            )
        except UnidentifiedImageError as exc:
            raise ScoValueError("Image signature invalide !") from exc

    return objects


def dict_pv_lettre_individuelle(
    formsemestre: FormSemestre, etud: Identite, decision: dict, legacy=True
) -> dict:
    """Renvoie un dictionnaire de paramètres pour la lettre individuelle
    (PV de jury).
    Si legacy (pour pdf), inclue l'identité, l'adresse, et d'autres champs, "à plat".
    Sinon, juste les champs texte pour les templates Jinja docx.
    """
    params = sco_bulletins.dict_infos_parcours_etud(etud.id, formsemestre)
    situation_etud: SituationEtudCursus = decision["Se"]
    titre_jury, titre_jury_court = jury_titres(
        formsemestre,
        situation_etud.parcours_validated() or not situation_etud.semestre_non_terminal,
    )
    params.update(
        {
            "semestre_id": formsemestre.semestre_id,
            "decision_sem_descr": decision["decision_sem_descr"],
            "type_jury": titre_jury,  # type de jury (passage ou delivrance)
            "type_jury_abbrv": titre_jury_court,  # idem, abbrégé
            "decisions_ue_descr": decision["decisions_ue_descr"],
            "prev_semestre_id": (
                decision["prev"]["semestre_id"] if decision["prev_decision_sem"] else ""
            ),
            "prev_decision_sem_txt": "",
            "decision_orig": "",
            "mention": decision["mention"],  # mention au diplôme
            # Informations sur compensations:
            "observation_txt": (
                f"""<b>Observation :</b> {decision["observation"]}."""
                if decision["observation"]
                else ""
            ),
            "INSTITUTION_CITY": (
                sco_preferences.get_preference("INSTITUTION_CITY", formsemestre.id)
                or ""
            ),
        }
    )
    # UE capitalisées:
    if decision["decisions_ue"] and decision["decisions_ue_descr"]:
        ue_plural = "s" if decision["decisions_ue_nb"] > 1 else ""
        params["decision_ue_txt"] = (
            f"""<b>Unité{ue_plural} d'Enseignement capitalisée{ue_plural
                } : {params["decisions_ue_descr"]}</b>"""
        )
    else:
        params["decision_ue_txt"] = ""

    # UE capitalisées avec ECTS
    if decision.get("decisions_ue_ects_txt"):
        ue_plural = "s" if decision["decisions_ue_nb"] > 1 else ""
        params[
            "decisions_ue_ects_txt"
        ] = f"""<b>Unité{ue_plural} d'Enseignement capitalisée{ue_plural
            } :</b> {decision["decisions_ue_ects_txt"]}"""
    else:
        params["decisions_ue_ects_txt"] = ""

    # Autorisations de passage
    if decision["autorisations"] and not situation_etud.parcours_validated():
        plural = "s" if len(decision["autorisations"]) > 1 else ""
        params[
            "autorisations_txt"
        ] = f"""Vous êtes autorisé{etud.e
            } à continuer dans le{plural} semestre{plural
            } : <b>{decision["autorisations_descr"]}</b>"""
    else:
        params["autorisations_txt"] = ""

    # Diplome:
    if (
        formsemestre.formation.is_apc() or decision["decision_sem"]
    ) and situation_etud.parcours_validated():
        params["diplome_txt"] = (
            f"""Vous avez donc obtenu le diplôme : <b>{formsemestre.formation.titre_officiel}</b>"""
        )
    else:
        params["diplome_txt"] = ""

    # Identité et adresse: (pour pdf et ancien html)
    if legacy:
        params.update(decision["identite"])
        # fix domicile (saut de lignes pour pdf ou html)
        if params.get("domicile"):
            params["domicile"] = params["domicile"].replace("\\n", "<br/>")

    # Les fonctions ci-dessous ajoutent ou modifient des champs:
    params["decision_ue_sans_niveaux_txt"] = params["decision_ue_txt"]
    if formsemestre.formation.is_apc():
        # ajout champs spécifiques PV BUT
        add_apc_infos(formsemestre, params, decision)
    else:
        # ajout champs spécifiques PV DUT
        add_classic_infos(formsemestre, params, decision)

    return params


def add_classic_infos(formsemestre: FormSemestre, params: dict, decision: dict):
    """Ajoute les champs pour les formations classiques, donc avec codes semestres"""
    if decision["prev_decision_sem"]:
        params["prev_code_descr"] = decision["prev_code_descr"]
        params[
            "prev_decision_sem_txt"
        ] = f"""<b>Décision du semestre antérieur S{params['prev_semestre_id']} :</b> {
            params['prev_code_descr']}"""
    # Décision semestre courant:
    if formsemestre.semestre_id >= 0:
        params["decision_orig"] = f"du semestre S{formsemestre.semestre_id}"
    else:
        params["decision_orig"] = ""


def add_apc_infos(formsemestre: FormSemestre, params: dict, decision: dict):
    """Ajoute à params les champs pour les formations APC (BUT), avec codes RCUE et année
    et diplome_dut120_descr
    """
    annee_but = (formsemestre.semestre_id + 1) // 2
    params["decision_orig"] = f"année BUT{annee_but}"
    if decision is None:
        params["decision_sem_descr"] = ""
        params["decision_ue_txt"] = ""
    else:
        decision_annee = decision.get("decision_annee") or {}
        params["decision_sem_descr"] = decision_annee.get("code") or ""
        niveaux_txt = (
            f"""<br/>&nbsp;&nbsp;&nbsp;&nbsp;- {
            '<br/>&nbsp;&nbsp;&nbsp;&nbsp;- '.join( decision.get("descr_decisions_rcue_list", []) )
            }"""
            if decision.get("descr_decisions_rcue_list")
            else "aucun"
        )
        params["niveaux_competences_txt"] = (
            f"""<b>Niveaux de compétences validés ce semestre :</b> {niveaux_txt}"""
        )
        params["decision_ue_txt"] = (
            f"""{params["decision_ue_txt"]}<br/>{params["niveaux_competences_txt"]}"""
        )
    add_dut120_infos(formsemestre, decision["identite"]["etudid"], params)


def add_dut120_infos(formsemestre: FormSemestre, etudid: int, params: dict):
    """Ajout DUT120:
    diplome_dut120_descr: phrase explicative si DUT enregistré
    diplome_dut120: booléen, vrai si DUT enregistré
    """
    validation_dut120 = ValidationDUT120.query.filter_by(
        etudid=etudid, formsemestre_id=formsemestre.id
    ).first()
    if validation_dut120:
        params["diplome_dut120"] = True
        params["diplome_dut120_descr"] = "Diplôme de DUT (BUT1, BUT2) validé"
        if "decision_sem_descr" in params:
            # sur decision_sem_descr afin que cela apparaisse sur les
            # lettres individuelles sans avoir à modifier le paramétrage
            params["decision_sem_descr"] += ". " + params["diplome_dut120_descr"]
    else:
        params["diplome_dut120"] = False
        params["diplome_dut120_descr"] = ""
