##############################################################################
#
# Gestion scolarite IUT
#
# Copyright (c) 1999 - 2025 Emmanuel Viennet.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Emmanuel Viennet      emmanuel.viennet@viennet.net
#
##############################################################################

"""Edition des lettres individuelles de jury au format docx"""

from html.parser import HTMLParser
from io import BytesIO

import docxtpl
from docx.shared import Mm
import jinja2

from app.models import FormSemestre, Identite
from app.scodoc import sco_preferences
from app.scodoc.sco_pv_lettres_individuelles import (
    dict_pv_lettre_individuelle,
    filter_decisions,
    get_lettres_param,
)
from app.scodoc import sco_pv_dict
from app.scodoc.sco_exceptions import ScoValueError


def docx_lettres_individuelles(
    formsemestre_id: int,
    etudids: list[int] | None = None,
    date_jury="",
    date_commission="",
    signature_img: bytes | None = None,
    template_docx: bytes = b"",
) -> bytes:
    """Document PDF avec les lettres d'avis pour les etudiants mentionnés
    (tous ceux du semestre, ou la liste indiquée par etudids).
    Si template=None, utilise template par défaut.
    Renvoie docx data ou vide si aucun étudiant avec décision de jury.
    """
    # Récupère toutes les décisions de jury pour le semestre
    dpv = sco_pv_dict.dict_pvjury(formsemestre_id, etudids=etudids, with_prev=True)
    if not dpv:
        return b""
    formsemestre = FormSemestre.get_formsemestre(formsemestre_id)
    params = get_lettres_param(
        formsemestre, date_jury=date_jury, date_commission=date_commission
    )
    convert_params_for_docx(params)
    decisions = filter_decisions(dpv["decisions"])
    if not decisions:
        return b""

    letter_template_stream = BytesIO(template_docx)
    # Données pour chaque étudiant:
    decisions_values = []
    for decision in decisions:
        etud = Identite.get_etud(decision["identite"]["etudid"])
        dec = {
            "etud": etud.to_dict_bul(),
            "jury": convert_params_for_docx(
                dict_pv_lettre_individuelle(formsemestre, etud, decision)
            ),
        }
        decisions_values.append(dec)

    doc = docxtpl.DocxTemplate(letter_template_stream)
    # Signature
    if signature_img:
        image_stream = BytesIO(signature_img)
        signature = docxtpl.InlineImage(doc, image_stream, width=Mm(50))
    else:
        signature = None
    try:
        doc.render(
            {
                "date_jury": date_jury,
                "date_commission": date_commission,
                # passer un dict et non l'objet pour empêcher l'appel non légitime de méthodes !
                "formsemestre": formsemestre.to_dict_api(),
                "titre_formation": formsemestre.formation.titre_officiel,
                "config": sco_preferences.SemPreferences(formsemestre.id),
                "decisions": decisions_values,
                "signature": signature if signature else None,
            },
            autoescape=True,
        )
    except jinja2.exceptions.TemplateError as e:
        raise ScoValueError(
            f"Erreur dans le template de lettre individuelle: {e}"
        ) from e

    # save to a bytes stringio
    output = BytesIO()
    doc.save(output)
    output.seek(0)
    return output.read()


class SimpleHTMLToDocxParser(HTMLParser):
    """Parser pour les champs générés avec du markup html
    (bold, italic, <br>) afin de les rendre correctement en Word
    """

    def __init__(self):
        super().__init__()
        self.rt = docxtpl.RichText()
        self.current_styles = {
            "bold": False,
            "italic": False,
        }

    def handle_starttag(self, tag, attrs):
        if tag in ("b", "strong"):
            self.current_styles["bold"] = True
        elif tag in ("i", "em"):
            self.current_styles["italic"] = True
        elif tag == "br":
            self.rt.add("\n")  # newline

    def handle_endtag(self, tag):
        if tag in ("b", "strong"):
            self.current_styles["bold"] = False
        elif tag in ("i", "em"):
            self.current_styles["italic"] = False

    def handle_data(self, data):
        self.rt.add(
            data, bold=self.current_styles["bold"], italic=self.current_styles["italic"]
        )

    def get_rich_text(self):
        return self.rt

    def reset(self):
        self.rt = docxtpl.RichText()
        return super().reset()


def convert_params_for_docx(params) -> dict:
    """Convertit les champs de params pour les rendre compatibles avec docx
    (RichText pour les champs avec du html)
    """
    parser = SimpleHTMLToDocxParser()

    for k in (
        "PV_LETTER_TEMPLATE",
        "bul_pdf_sig_right",
        "INSTITUTION_NAME",
        "bul_but_pdf_title",
        "bul_pdf_caption",
        "bul_pdf_title",
        "bul_pdf_sig_left",
        "PV_INTRO",
        "PV_LETTER_PASSAGE_SIGNATURE",
        "PV_LETTER_DIPLOMA_SIGNATURE",
        "INSTITUTION_ADDRESS",
        "decision_ue_txt",
        "decisions_ue_ects_txt",
        "autorisations_txt",
        "decision_ue_sans_niveaux_txt",
        "niveaux_competences_txt",
    ):
        v = params.get(k)
        if isinstance(v, str):
            parser.feed(v)
            params[k] = parser.get_rich_text()
            parser.reset()
    return params
