// Procedures communes pour formulaires saisie des notes
/***************************************************
 Pré-requis pour l'utilisation des fonctionnalités communes à saisie_notes_par_etu et saisie_notes_par_eval
 * il existe un champ d'id `formnotes_formsemestre_id` qui donne l'id du formsemestre courant
 * Les champs de saisie de notes:
     * répondent au sélecteur ("#formnotes .note")
     * possèdent un id de la forme `formnotes_note_###` (### identifiant étudiant, évaluation ou autre)
     * possèdent les attributs initialisés `data-last-saved-value`, `data-original-value` `data_modified` = false;
 * Les champs historiques:
     * possèdent un id de la forme `hist_###` (### identifiant étudiant, évaluation ou autre)
 * les fonctions javascript suivantes sont implémentées:
     * get_id(field): str (retourne le ### pour un champ note donné)
     * get_evaluation_id(identifiant: str): int
     * get formsemestre_id(identifiant: str): int
     * get_note_min(identifiant: str): int
     * get_note_max(identifiant: str): int
     * update_jurylink(note_field: element, etudid: int)

 Fonctionnement:
     Un identifiant (field_id) represente un chaîne caractéristique d'une ligne de saisie.
         * etudid pour le formulaire de saisie par évaluation
         * evaluation_id pour le formulaire de saisie par étudiant
     un attribut data-modified inque si un changement a eu lieu sur une zone de saisie.
 initialement
     l attribut data-modified indique toute modification
     dans le traitement on_blur on vérifie si la valeur a été changée ou pas
     exemple '15' changée en '15.0'
        input => data-modifed = true
        blur => rien n'est fait (la valeur n est pas changée)
 modification (JMP)
     exemple '15' changée en '15.0'
        input => data-modified = false
 *****************************************************/


const note_prefix = "formnotes_note_";
const hist_prefix = "formnotes_hist_";

function get_formsemestre_id() {
  return document.getElementById("formnotes_formsemestre_id").getAttribute("value");;
}

let nbSaving = 0; // nombre de requêtes en cours

function incSaving() {
  nbSaving++;
  // disable button 'Terminer'
  let anchor = document.querySelector(".link-terminer")
  anchor.style.pointerEvents = "none";
  anchor.style.color = "gray";
  anchor.style.textDecoration = "none";
  anchor.style.cursor = "not-allowed";
}

function decSaving() {
  nbSaving--;
  if (nbSaving <= 0) {
    // re-enable bouton 'Terminer'
    let anchor = document.querySelector(".link-terminer");
    anchor.style.pointerEvents = "unset";
    anchor.style.color = "unset";
    anchor.style.textDecoration = "unset";
    anchor.style.cursor = "unset";
  }
}

window.addEventListener('beforeunload', function (e) {
  const noteInputs = document.querySelectorAll("#formnotes .note");
  noteInputs.forEach(function (input) {
    if (input.getAttribute("data-modified") === "true" && input.value !== input.getAttribute("data-last-saved-value")) {
      valid_note.call(input);
    }
  });
  if (nbSaving > 0) {
    // Display a confirmation dialog
    const confirmationMessage = 'Des modifications sont en cours de sauvegarde. Êtes-vous sûr de vouloir quitter cette page ?';
    e.preventDefault(); // Standard for most modern browsers
    e.returnValue = confirmationMessage; // For compatibility with older browsers
    return confirmationMessage; // For compatibility with older browsers
  }
});

function is_valid_note(field_id, v) {
  if (!v) return true;

  if (!v.match("^-?[0-9]*.?[0-9]*$")) {
    return v == "ABS" || v == "EXC" || v == "SUPR" || v == "ATT" || v == "DEM";
  } else {
    const x = parseFloat(v);
    return x >= get_note_min(field_id) && x <= get_note_max(field_id);
  }
}

function valid_note() {
  const field_id = get_field_id(this)
  const v = this.value.trim().toUpperCase().replace(",", ".");
  if (is_valid_note(field_id, v)) {
    if (v && v != this.getAttribute("data-last-saved-value")) {
      this.className = "note_valid_new";
      const etudid = get_etudid(field_id);
      save_note(field_id, v);
    }
  } else {
    /* Saisie invalide */
    this.className = "note_invalid";
    sco_message("valeur invalide ou hors barème");
  }
}

function get_field_id(elt) {      // la forme des id est ????_#### (où #### est le  field_id que l'on cherche)
  return elt.id.split('_').pop(); // formnotes_note_#### ou eval_####
}

function get_hist_field(field_id) {
  return document.getElementById(hist_prefix + field_id);
}

function get_note_field(field_id) {
  return document.getElementById(note_prefix + field_id);
}

function write_on_blur(elt) {
    if (elt.getAttribute("data-modified") === "true") {
        if (elt.value !== elt.getAttribute("data-last-saved-value")) {
            valid_note.call(elt);
            elt.setAttribute("data-modified", "false"); // Reset the modified flag
        }
    }
}

async function save_note(field_id, v) {
  const evaluation_id = get_evaluation_id(field_id);
  const etudid = get_etudid(field_id);
  const scoMsg = document.getElementById("sco_msg");
  const note_field = get_note_field(field_id);
  const hist_field = get_hist_field(field_id);
  scoMsg.innerHTML = "en cours...";
  scoMsg.style.display = "block";
  incSaving(); // update counter to show one more saving in progress
  try {
    const response = await fetch(
      SCO_URL + "../api/evaluation/" + evaluation_id + "/notes/set",
      {
        method: "POST",
        headers: {
          "Content-Type": "application/json",
        },
        body: JSON.stringify({
          notes: [[etudid, v]],
          comment: document.getElementById("formnotes_comment").value,
        }),
      }
    );
    if (!response.ok) {
      sco_message("Erreur: valeur non enregistrée");
    } else {
      const data = await response.json();
      scoMsg.style.display = "none";
      if (data.etudids_changed.length > 0) {
        sco_message("enregistré");
        note_field.className = "note_saved";
        // Il y avait une decision de jury ?
        if (data.etudids_with_decision.includes(etudid)) {
          update_jurylink(field_id, v);
        }
        // Mise à jour menu historique
        if (data.history_menu[etudid]) {
          hist_field.innerHTML = data.history_menu[etudid];
        }
        note_field.setAttribute("data-last-saved-value", v);
      }
    }
  } catch (error) {
    console.error("Fetch error:", error);
    sco_message("Erreur réseau: valeur non enregistrée");
  } finally {
    decSaving(); // mise à jour du nombre de requêtes en instance. si devient nul, réactive le lien 'Terminer'
  }
}

function make_jurylink(field_id) {
    const formsemestre_id = get_formsemestre_id();
    const etudid = get_etudid(field_id);
    const href = `${SCO_URL}Notes/formsemestre_validation_etud_form?formsemestre_id=${formsemestre_id}&etudid=${etudid}`;
    return '<a href="' + href +'">mettre à jour décision de jury</a>';
}

function change_history(e) {
  const opt = e.selectedOptions[0];
  const val = opt.getAttribute("data-note");
  if (val != '') {
      const field_id = get_field_id(e.parentElement);
      // le input associé a ce menu:
      const input = get_note_field(field_id)
      save_note(field_id, val);
      input.value = val
      // gère le style de l'input (note_saved ou pas selon que la valeur de la note a changé par rapport à la valeur initiale)
      if (val == input.getAttribute("data-orig-value")) {
          input.classList.remove("note_saved")
      } else {
          input.classList.add("note_saved");
      }
  }
}

