##############################################################################
# ScoDoc
# Copyright (c) 1999 - 2025 Emmanuel Viennet.  All rights reserved.
# See LICENSE
##############################################################################

"""Table bilan des étudiants des semestres pour AutoSco:
semblable à export_etudiants_courants
mais avec colonnes supplémentaires:

- profil (étudiant, personnel, extérieur)
- nom, prénom, INE, NIP, e-mail, téléphone
- cours (titre du semestre)
- moyenne générale
- nombre de présences
- codes Apogée de l'étudiant (si inscrit ailleurs)
- enseignant
"""
from flask import g

from app.auth.models import User
from app.comp import res_sem
from app.models import FormSemestre
import app.scodoc.sco_assiduites as scass
from app.scodoc import sco_preferences
from app.scodoc import sco_utils as scu
from app.scodoc.sco_exceptions import ScoValueError
from app.tables.list_etuds import TableEtudWithInfos, RowEtudWithInfos


class TableBilanAutoSco(TableEtudWithInfos):
    """Table listant des étudiants avec bilan AutoSco"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.assi_metric = scu.translate_assiduites_metric(
            sco_preferences.get_preference("assi_metrique", dept_id=g.scodoc_dept_id),
        )

    def add_formsemestre(self, formsemestre: FormSemestre):
        "Ajoute les étudiants de ce semestre à la table"
        # Mémorise inscriptions pour optimiser accès
        self.inscriptions = formsemestre.etuds_inscriptions
        super().add_formsemestre(formsemestre)


class RowEtudWithAutoSco(RowEtudWithInfos):
    """Ligne de la table d'étudiants avec colonnes UEs"""

    with_boursier = False  # inutile dans cette table
    with_departement = False
    with_formsemestre = True  # pour le titre du semestre
    with_ine = True

    def add_etud_cols(self):
        """Ajoute colonnes pour cet étudiant"""
        super().add_etud_cols()
        etud = self.etud
        # Ajoute les colonnes AutoSco
        self.add_cell(
            "email",
            "email",
            etud.adresses.first().email,
            "identite_detail",
            # target=self.target_url,
            # target_attrs={"class": "discretelink", "title": self.target_title},
        )
        self.add_cell(
            "telephone", "telephone", etud.adresses.first().telephone, "identite_detail"
        )
        # Moyenne générale
        nt = res_sem.load_formsemestre_results(self.formsemestre)
        moy_gen = nt.get_etud_moy_gen(etud.id)
        self.add_cell(
            "moy_gen",
            "Moyenne",
            scu.fmt_note(moy_gen),
            "resultats",
            raw_content=moy_gen,
        )
        # Nombre de présences
        compte_etat: dict[str, dict] = scass.get_assiduites_stats(
            assiduites=etud.assiduites,
            metric=self.table.assi_metric,
            filtered={
                "date_debut": self.formsemestre.date_debut,
                "date_fin": self.formsemestre.date_fin,
                "etat": "present",  # uniquement les présences
                # "formsemestre_modimpls": formsemestre, # restreint à ce formsemestre,
                # "split": 1,  # afin d'avoir la division des stats en état, etatjust, etatnonjust
            },
        )
        nb_presences = compte_etat.get(self.table.assi_metric, "-")
        self.add_cell(
            "nb_presences",
            "Présences",
            str(nb_presences),
            "resultats",
            raw_content=nb_presences,
        )
        # Enseignant : celui du premier module, ou a defaut le 1er resp. du semestre
        modimpl = self.formsemestre.modimpls.first()
        enseignant: User = (
            modimpl.responsable
            if modimpl
            else (self.formsemestre.responsables or [None])[0]
        )
        self.add_cell(
            "enseignant",
            "Enseignant 1",
            enseignant.get_prenomnom() if enseignant else "-",
            "semestre",
            data={"order": enseignant.sort_key() if enseignant else "-"},
        )
        # Code Apogée de l'inscription (/!\ pas celui du semestre)
        # utilisé par AuoSco pour indiquer l'inscription à la formation principale de l'étudiant
        inscription = self.table.inscriptions.get(etud.id)
        self.add_cell(
            "etape",
            "Étape étudiant",
            inscription.etape if inscription else "-",
            "semestre",
        )


def table_etudiants_courants(formsemestres: list[FormSemestre]) -> TableBilanAutoSco:
    """Table des étudiants des formsemestres indiqués"""
    if not formsemestres:
        raise ScoValueError("Aucun semestre en cours")
    table = TableBilanAutoSco(row_class=RowEtudWithAutoSco)
    for formsemestre in formsemestres:
        table.add_formsemestre(formsemestre)
    return table
