"""Edition et liste des templates docx"""

from flask import abort, flash, g, redirect, request, url_for
from flask import render_template
from flask_login import current_user
from werkzeug.utils import secure_filename

from app import db, log
from app.decorators import (
    scodoc,
    permission_required,
)
from app.forms.docx_template.edit_docx_template import CreateDocxTemplateForm
from app.models import DocxTemplate
from app.scodoc.sco_exceptions import ScoPermissionDenied, ScoValueError
from app.scodoc.sco_permissions import Permission
from app.scodoc import sco_utils as scu
from app.views import scolar_bp as bp, scodoc_bp


# Certaines routes sont partagées entre les deux blueprints
# globale (scodoc_bp) pour les templates communs à tous les départements
# et départementale (bp == scolar_bp) pour les templates spécifiques à un département.


@bp.route("/list_docx_templates/<string:template_type>")
@scodoc_bp.route("/list_docx_templates/<string:template_type>")
@scodoc
@permission_required(Permission.ScoView)
def list_docx_templates(template_type):
    "List all available templates of given type for the current dept"
    if template_type not in DocxTemplate.template_type.type.enums:
        raise ScoValueError("Template type non valide")
    global_templates, dept_templates = DocxTemplate.get_dept_templates(
        dept_id=g.scodoc_dept_id, template_type=template_type
    )
    template_type_str = DocxTemplate.get_template_type_str(template_type)
    return render_template(
        "docx/list_templates.j2",
        dept_templates=dept_templates,
        global_templates=global_templates,
        dept_id=g.scodoc_dept_id,
        template_type=template_type,
        template_type_str=template_type_str,
        title=f"Templates docx {template_type_str}",
    )


@bp.route("/delete_docx_template/<int:template_id>", methods=["GET", "POST"])
@scodoc_bp.route("/delete_docx_template/<int:template_id>", methods=["GET", "POST"])
@scodoc
@permission_required(Permission.EditPreferences)
def delete_docx_template(template_id: int):
    "Supprime un template docx (avec confirmation si GET)"
    template = db.session.get(DocxTemplate, template_id)
    if not template:
        abort(404, "Template non trouvé")
    if not template.can_be_edited_by():
        raise ScoPermissionDenied(
            dest_url=url_for("scolar.list_docx_templates", scodoc_dept=g.scodoc_dept)
        )

    dest_url = (
        url_for(
            "scolar.list_docx_templates",
            template_type=template.template_type,
            scodoc_dept=g.scodoc_dept,
        )
        if g.scodoc_dept
        else url_for("scodoc.list_docx_templates", template_type=template.template_type)
    )
    # On supprime le template
    if request.method == "POST":
        db.session.delete(template)
        db.session.commit()
        flash("Template supprimé", "success")
        return redirect(dest_url)
    return render_template(
        "confirm_dialog.j2",
        title=f"""Effacer le template doxc {template.name} (pour {template.template_type_str()}) ?""",
        cancel_url=dest_url,
    )


@bp.route("/edit_docx_template/<int:template_id>", methods=["GET", "POST"])
@scodoc_bp.route("/edit_docx_template/<int:template_id>", methods=["GET", "POST"])
@scodoc
@permission_required(Permission.EditPreferences)
def edit_docx_template(template_id: int):
    """Edit an existing Docx template."""
    template = db.session.get(DocxTemplate, template_id)
    if not template:
        abort(404, "Template non trouvé")
    if not template.can_be_edited_by():
        raise ScoPermissionDenied(
            dest_url=url_for("scolar.list_docx_templates", scodoc_dept=g.scodoc_dept)
        )
    form = CreateDocxTemplateForm(
        name=template.name, description=template.description, edit=True
    )
    if form.validate_on_submit():
        data = form.data.copy()
        if form.docx_file.data:  # Modif du fichier
            data["name"] = form.name.data or secure_filename(
                form.docx_file.data.filename
            )
            data["template"] = form.docx_file.data.read()
        modified = template.from_dict(data)
        db.session.commit()
        flash("Template mis à jour avec succès", "success")
        log(f"edit_docx_template: {template}")
        log(f"modified={modified}")
        return redirect(
            url_for(
                "scolar.list_docx_templates",
                template_type=template.template_type,
                scodoc_dept=g.scodoc_dept,
            )
            if g.scodoc_dept
            else url_for(
                "scodoc.list_docx_templates", template_type=template.template_type
            )
        )
    return render_template(
        "docx/edit_docx_template.j2",
        form=form,
        template=template,
        template_type=template.template_type,
        title=f"Modifier le template docx pour les {template.template_type_str()}",
    )


@bp.route("/create_docx_template/<string:template_type>", methods=["GET", "POST"])
@scodoc_bp.route(
    "/create_docx_template/<string:template_type>", methods=["GET", "POST"]
)
@scodoc
@permission_required(Permission.EditPreferences)
def create_docx_template(template_type: str):
    """Create a new Docx template of the given template_type,
    by uploading a docx file.
    The name and description are given in the form.
    author, creation date and modification date are set automatically.
    """
    # Check if the template_type is valid
    if template_type not in DocxTemplate.template_type.type.enums:
        raise ScoValueError("Template type non valide")
    form = CreateDocxTemplateForm()
    if form.validate_on_submit():
        # Create a new DocxTemplate object
        name = form.name.data or secure_filename(form.docx_file.data.filename)
        template = DocxTemplate(
            name=name,
            description=form.description.data,
            template=form.docx_file.data.read(),
            template_type=template_type,
            dept_id=g.scodoc_dept_id if g.scodoc_dept else None,  # None si global
            author_id=current_user.id,
        )
        db.session.add(template)
        db.session.commit()
        log(f"create_docx_template: {template}")
        flash("Template créé avec succès", "success")
        return redirect(
            url_for(
                "scolar.list_docx_templates",
                template_type=template_type,
                scodoc_dept=g.scodoc_dept,
            )
            if g.scodoc_dept
            else url_for("scodoc.list_docx_templates", template_type=template_type)
        )
    return render_template(
        "docx/edit_docx_template.j2",
        form=form,
        title="Créer un nouveau template",
        template_type=template_type,
    )


@bp.route("/download_docx_template/<int:template_id>")
@scodoc_bp.route("/download_docx_template/<int:template_id>")
@scodoc
def download_docx_template(template_id: int):
    """Download Docx template (.docx document)."""
    template: DocxTemplate = DocxTemplate.get_instance(template_id)
    if not template.can_be_viewed_by():
        raise ScoPermissionDenied(dest_url=url_for("scodoc.index"))

    return scu.send_file(
        template.template,
        filename=template.name,
        suffix=".docx" if not template.name.endswith(".docx") else "",
        mime=scu.DOCX_MIMETYPE,
        attached=True,
    )
