"""
Génération des lettres individuelles PV
"""

import time

from flask import flash, g, redirect, render_template, request, url_for

from app.decorators import scodoc, permission_required
from app.forms.formsemestre.gen_lettres_inviduelles import (
    GenerateLettresIndividuellesForm,
)
from app.models import DocxTemplate, FormSemestre
from app.scodoc import sco_groups, sco_groups_view, sco_pv_lettres_individuelles_docx
from app.scodoc.sco_exceptions import ScoValueError
from app.scodoc.sco_permissions import Permission
from app.scodoc import sco_utils as scu
from app.views import notes_bp as bp
from app.views import ScoData


@bp.route(
    "/formsemestre/<int:formsemestre_id>/lettres_individuelles/docx",
    methods=["GET", "POST"],
)
@scodoc
@permission_required(Permission.EditFormSemestre)
def formsemestre_lettres_individuelles_docx(formsemestre_id: int):
    """Génération des lettres individuelles en format docx"""
    formsemestre: FormSemestre = FormSemestre.get_or_404(formsemestre_id)
    if request.method == "POST":
        group_ids = request.form.getlist("group_ids")
    else:
        group_ids = request.args.getlist("group_ids")
    try:
        group_ids = [int(gid) for gid in group_ids]
    except ValueError as exc:
        raise ScoValueError("group_ids invalide") from exc
    if not group_ids:
        # tous les inscrits du semestre
        group_ids = [sco_groups.get_default_group(formsemestre_id)]
    groups_infos = sco_groups_view.DisplayedGroupsInfos(
        group_ids, formsemestre_id=formsemestre_id
    )
    etudids = [m["etudid"] for m in groups_infos.members]

    form = GenerateLettresIndividuellesForm()
    global_templates, dept_templates = DocxTemplate.get_dept_templates(
        dept_id=g.scodoc_dept_id, template_type="lettre_individuelle"
    )
    form.template_select.choices = (
        [(-1, "Standard ScoDoc")]
        + [(t.id, f"{t.name} (global)") for t in global_templates]
        + [(t.id, f"{t.name} ({t.departement.acronym})") for t in dept_templates]
    )
    if form.validate_on_submit():
        template_docx = DocxTemplate.get_template_docx(
            form.template_select.data, "lettre_individuelle"
        )
        doc = sco_pv_lettres_individuelles_docx.docx_lettres_individuelles(
            formsemestre.id,
            etudids=etudids,
            date_jury=form.date_jury.data,
            date_commission=form.date_commission.data,
            template_docx=template_docx,
            signature_img=(
                form.signature_img.data.read() if form.signature_img.data else None
            ),
        )
        if not doc:
            flash("Aucun étudiant n'a de décision de jury !")
            return redirect(
                url_for(
                    "notes.formsemestre_status",
                    scodoc_dept=g.scodoc_dept,
                    formsemestre_id=formsemestre_id,
                )
            )
        groups_filename = "-" + groups_infos.groups_filename
        filename = f"""lettres-{formsemestre.titre_num()}{groups_filename}-{
            time.strftime("%Y-%m-%d")}"""
        mime, suffix = scu.get_mime_suffix("docx")
        flash("document généré")
        return scu.send_file(doc, filename=filename, mime=mime, suffix=suffix)

    return render_template(
        "formsemestre/lettres_individuelles_docx.j2",
        form=form,
        formsemestre=formsemestre,
        sco=ScoData(formsemestre=formsemestre),
        title="Génération des lettres docx",
    )
