"""Test formsemestre

Utilisation :
    créer les variables d'environnement: (indiquer les valeurs
    pour le serveur ScoDoc que vous voulez interroger)

    export SCODOC_URL="https://scodoc.xxx.net/"
    export SCODOC_USER="xxx"
    export SCODOC_PASSWD="xxx"
    export CHECK_CERTIFICATE=0 # ou 1 si serveur de production avec certif SSL valide

    (on peut aussi placer ces valeurs dans un fichier .env du répertoire tests/api).

    Lancer (i un serveur est déjà lancé):
        pytest tests/api/test_api_formsemestre_demande_inscr.py
    ou bien:
        tools/test_api.sh -x --pdb tests/api/test_api_formsemestre_demande_inscr.py
"""

import base64

from tests.api.setup_test_api import (
    APIError,
    api_admin_headers,
    DEPT_ACRONYM,
    GET,
    POST,
    set_headers,
)


def test_formsemestre_demande_inscr(api_admin_headers):
    """
    Routes:
        - `formsemestre/<int:formsemestre_id>/demande_inscription/create`
        - `formsemestre/<int:formsemestre_id>/demande_inscription/<str:demande_num>/edit`
        - `formsemestre/<int:formsemestre_id>/demande_inscription/<str:demande_num>/delete`
        - `formsemestre/<int:formsemestre_id>/demande_inscription/list`
        - `formsemestre/<int:formsemestre_id>/demande_inscription/<str:demande_num>`
        - `formsemestre/<int:formsemestre_id>/demande_inscription/<str:demande_num>/piece_jointe`
    """
    set_headers(api_admin_headers)
    # Create a formsemestre
    formations_ids = GET("/formations_ids", dept=DEPT_ACRONYM)
    assert formations_ids, "No formations found."
    formation_id = formations_ids[0]
    data = {
        "formation_id": formation_id,
        "date_debut": "2025-01-02",
        "date_fin": "2025-06-30",
        "modalite": "FI",
        "semestre_id": 1,
        "titre": "Semestre 1 test demande inscription",
    }
    formsemestre = POST("/formsemestre/create", dept=DEPT_ACRONYM, data=data)
    assert formsemestre, "No formsemestre created."
    assert formsemestre["formation_id"] == formation_id
    formsemestre_id = formsemestre["id"]

    # Create description
    description = POST(
        f"/formsemestre/{formsemestre_id}/description/edit",
        data={
            "date_debut_inscriptions": "2025-01-02T00:00:00+01:00",
            "date_fin_inscriptions": "2666-12-31T00:00:00+01:00",
        },
    )
    assert description["formsemestre_id"] == formsemestre_id
    assert description["date_debut_inscriptions"] == "2025-01-02T00:00:00+01:00"
    assert description["date_fin_inscriptions"] == "2666-12-31T00:00:00+01:00"
    # Get a test image
    with open("tests/ressources/images/papillon.jpg", "rb") as f:
        image = f.read()

    # Create a demande_inscription
    motivation = "I want to enroll in this semester."
    data = {
        "nom": "Doe",
        "prenom": "John",
        "email": "john.doe@example.com",
        "telephone": "0123456789",
        "niveau": "debutant",
        "motivation": motivation,
        "piece_jointe": base64.b64encode(image).decode("utf-8"),
    }
    demande = POST(
        f"/formsemestre/{formsemestre_id}/demande_inscription/create", data=data
    )
    numero = demande["numero"]
    assert numero
    assert isinstance(numero, str)
    assert demande["motivation"] == motivation
    #
    demande_ret = GET(f"/formsemestre/{formsemestre_id}/demande_inscription/{numero}")
    assert base64.b64decode(demande_ret["piece_jointe"]) == image
    assert demande_ret["motivation"] == motivation
    # Create another one
    data["nom"] = "Jane"
    del data["piece_jointe"]
    demande2 = POST(
        f"/formsemestre/{formsemestre_id}/demande_inscription/create", data=data
    )
    assert demande2["numero"]
    # List all demandes_inscriptions
    demandes = GET(f"/formsemestre/{formsemestre_id}/demandes_inscriptions")
    assert len(demandes["data"]) == 2
    # Check order: plus ancienne d'abord
    assert demandes["data"][0]["date_demande"] < demandes["data"][1]["date_demande"]
    # Edit the first one: change motivation
    new_motivation = "I want to enroll in this semester. Please."
    demande_ret = POST(
        f"/formsemestre/{formsemestre_id}/demande_inscription/{numero}/edit",
        data={"motivation": new_motivation},
    )
    assert demande_ret["motivation"] == new_motivation
    # Edit: try to change statut to "accepté": impossible
    try:
        POST(
            f"/formsemestre/{formsemestre_id}/demande_inscription/{numero}/edit",
            data={"statut": "ACCEPTEE"},
        )
        raise AssertionError("Error: statut should not be editable by the user.")
    except APIError:
        pass
    # List the first one and verify content
    demande_ret = GET(f"/formsemestre/{formsemestre_id}/demande_inscription/{numero}")
    assert demande_ret["motivation"] == new_motivation
    assert demande_ret["numero"] == numero
    # Delete the second one
    ans = POST(
        f"/formsemestre/{formsemestre_id}/demande_inscription/{demande2['numero']}/delete",
        raw=True,
    )
    assert ans.status_code == 204, "Error deleting demande_inscription"

    # List all demandes_inscriptions
    demandes = GET(f"/formsemestre/{formsemestre_id}/demandes_inscriptions")
    assert len(demandes["data"]) == 1
    assert demandes["data"][0]["numero"] == numero
    # Remove image data
    demande_ret = POST(
        f"/formsemestre/{formsemestre_id}/demande_inscription/{numero}/edit",
        data={"piece_jointe": None},
    )
    assert demande_ret["nom"] == "Doe"
    demande_ret = GET(f"/formsemestre/{formsemestre_id}/demande_inscription/{numero}")
    assert demande_ret["piece_jointe"] is None

    # Delete formsemestre
    ans = POST(f"/formsemestre/{formsemestre_id}/delete", raw=True)
    assert ans.status_code == 204, f"Error deleting formsemestre {formsemestre_id}"
